"""
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License").
You may not use this file except in compliance with the License.
A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and limitations under the License.
"""

import re
import six
import yaml
from yaml.emitter import Emitter, ScalarAnalysis

from .literal import LiteralString
from .odict import ODict

TAG_MAP = "tag:yaml.org,2002:map"
TAG_STRING = "tag:yaml.org,2002:str"
AWS_ACCOUNT_ID = r"^0[0-9]+$"


class CfnEmitter(Emitter):
    def analyze_scalar(self, scalar):
        # We have Json payloads that we want to show as literal
        # This way we can skip the checks that analize_scalar do in the string when you have
        # leading_space or leading_break
        if not scalar or isinstance(scalar, LiteralString):
            return ScalarAnalysis(scalar=scalar, empty=True, multiline=False,
                                  allow_flow_plain=False, allow_block_plain=True,
                                  allow_single_quoted=True, allow_double_quoted=True,
                                  allow_block=True)
        return super(CfnEmitter, self).analyze_scalar(scalar)


class CfnYamlDumper(yaml.Dumper, CfnEmitter):
    """
    Indent block sequences from parent using more common style
    ("  - entry"  vs "- entry").
    Causes fewer problems with validation and tools.
    """

    def increase_indent(self, flow=False, indentless=False):
        return super(CfnYamlDumper, self).increase_indent(flow, False)

    def represent_scalar(self, tag, value, style=None):
        if re.match(AWS_ACCOUNT_ID, value):
            style = "\'"

        if isinstance(value, six.text_type):
            if any(eol in value for eol in "\n\r") and style is None:
                style = "\""

            # return super(CfnYamlDumper, self).represent_scalar(TAG_STRING, value, style)

        return super(CfnYamlDumper, self).represent_scalar(tag, value, style)


def string_representer(dumper, value):
    style = None

    if "\n" in value:
        style = "\""

    return dumper.represent_scalar(TAG_STRING, value, style=style)


def map_representer(dumper, value):
    """
    Map ODict into ODict to prevent sorting
    """

    return dumper.represent_mapping(TAG_MAP, value)


def literal_unicode_representer(dumper, value):
    return dumper.represent_scalar(TAG_STRING, value, style='|')


# Customise the dumper
CfnYamlDumper.add_representer(ODict, map_representer)
CfnYamlDumper.add_representer(LiteralString, literal_unicode_representer)
CfnYamlDumper.add_representer(six.text_type, string_representer)



"""
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file except in compliance with the License. A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
"""

from .odict import ODict
import six
import yaml

TAG_MAP = "tag:yaml.org,2002:map"
UNCONVERTED_SUFFIXES = ["Ref", "Condition"]
FN_PREFIX = "Fn::"


class CfnYamlLoader(yaml.Loader):
    pass


def multi_constructor(loader, tag_suffix, node):
    """
    Deal with !Ref style function format
    """

    if tag_suffix not in UNCONVERTED_SUFFIXES:
        tag_suffix = "{}{}".format(FN_PREFIX, tag_suffix)

    constructor = None

    if tag_suffix == "Fn::GetAtt":
        constructor = construct_getatt
    elif isinstance(node, yaml.ScalarNode):
        constructor = loader.construct_scalar
    elif isinstance(node, yaml.SequenceNode):
        constructor = loader.construct_sequence
    elif isinstance(node, yaml.MappingNode):
        constructor = loader.construct_mapping
    else:
        raise Exception("Bad tag: !{}".format(tag_suffix))

    return ODict((
        (tag_suffix, constructor(node)),
    ))


def construct_getatt(node):
    """
    Reconstruct !GetAtt into a list
    """

    if isinstance(node.value, six.text_type):
        return node.value.split(".", 1)
    elif isinstance(node.value, list):
        return [s.value for s in node.value]
    else:
        raise ValueError("Unexpected node type: {}".format(type(node.value)))


def construct_mapping(self, node, deep=False):
    """
    Use ODict for maps
    """

    mapping = ODict()

    for key_node, value_node in node.value:
        key = self.construct_object(key_node, deep=deep)
        value = self.construct_object(value_node, deep=deep)

        mapping[key] = value

    return mapping


# Customise our loader
CfnYamlLoader.add_constructor(TAG_MAP, construct_mapping)
CfnYamlLoader.add_multi_constructor("!", multi_constructor)



"""
Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file except in compliance with the License. A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
"""

from .json_encoder import DateTimeAwareJsonEncoder
from .odict import ODict
from .yaml_dumper import CfnYamlDumper
from .yaml_loader import CfnYamlLoader
import json
import yaml


def load_json(source):
    return json.loads(source, object_pairs_hook=ODict)


def dump_json(source):
    return json.dumps(source, indent=4, cls=DateTimeAwareJsonEncoder,
                      separators=(',', ': '), ensure_ascii=False)


def load_yaml(source):
    return yaml.load(source, Loader=CfnYamlLoader)


def dump_yaml(source):
    return yaml.dump(source, Dumper=CfnYamlDumper, default_flow_style=False, allow_unicode=True)



"""
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file except in compliance with the License. A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
"""

import collections


class OdictItems(list):
    """
    Helper class to ensure ordering is preserved
    """

    def __init__(self, items):
        new_items = []

        for item in items:
            class C(type(item)):
                def __lt__(self, *args, **kwargs):
                    return False

            new_items.append(C(item))

        return super(OdictItems, self).__init__(new_items)

    def sort(self):
        pass


class ODict(collections.OrderedDict):
    """
    A wrapper for OrderedDict that doesn't allow sorting of keys
    """

    def __init__(self, pairs=[]):
        if isinstance(pairs, dict):
            # Dicts lose ordering in python<3.6 so disallow them
            raise Exception("ODict does not allow construction from a dict")

        super(ODict, self).__init__(pairs)

        old_items = self.items
        self.items = lambda: OdictItems(old_items())



"""
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License").
You may not use this file except in compliance with the License. A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and limitations under the License.
"""

from . import flip
import click
import sys


@click.command()
@click.option("--input", "-i", "in_format", type=click.Choice(["json", "yaml"]), help="Specify the input format. Overrides -j and -y flags.")
@click.option("--output", "-o", "out_format", type=click.Choice(["json", "yaml"]), help="Specify the output format. Overrides -j, -y, and -n flags.")
@click.option("--json", "-j", "out_flag", flag_value="json", help="Convert to JSON. Assume the input is YAML.")
@click.option("--yaml", "-y", "out_flag", flag_value="yaml", help="Convert to YAML. Assume the input is JSON.")
@click.option("--clean", "-c", is_flag=True, help="Performs some opinionated cleanup on your template.")
@click.option("--long", "-l", is_flag=True, help="Use long-form syntax for functions when converting to YAML.")
@click.option("--no-flip", "-n", is_flag=True, help="Perform other operations but do not flip the output format.")
@click.argument("input", type=click.File("r"), default=sys.stdin)
@click.argument("output", type=click.File("w"), default=sys.stdout)
@click.version_option(message='AWS Cloudformation Template Flip, Version %(version)s')
@click.pass_context
def main(ctx, **kwargs):
    """
    AWS CloudFormation Template Flip is a tool that converts
    AWS CloudFormation templates between JSON and YAML formats,
    making use of the YAML format's short function syntax where possible.
    """
    in_format = kwargs.pop('in_format')
    out_format = kwargs.pop('out_format') or kwargs.pop('out_flag')
    no_flip = kwargs.pop('no_flip')
    clean = kwargs.pop('clean')
    long_form = kwargs.pop('long')
    input_file = kwargs.pop('input')
    output_file = kwargs.pop('output')

    if not in_format:
        if input_file.name.endswith(".json"):
            in_format = "json"
        elif input_file.name.endswith(".yaml") or input_file.name.endswith(".yml"):
            in_format = "yaml"

    if input_file.name == "<stdin>" and sys.stdin.isatty():
        click.echo(ctx.get_help())
        ctx.exit()

    try:
        output_file.write(flip(
            input_file.read(),
            in_format=in_format,
            out_format=out_format,
            clean_up=clean,
            no_flip=no_flip,
            long_form=long_form
        ))
    except Exception as e:
        raise click.ClickException("{}".format(e))



"""
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License").
You may not use this file except in compliance with the License. A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and limitations under the License.
"""

from .yaml_dumper import get_dumper
from cfn_clean import clean, cfn_literal_parser
from cfn_tools import load_json, load_yaml, dump_json
import yaml
import sys


def load(template):
    """
    Try to guess the input format
    """

    try:
        data = load_json(template)
        return data, "json"
    except ValueError as e:
        try:
            data = load_yaml(template)
            return data, "yaml"
        except Exception:
            raise e


def dump_yaml(data, clean_up=False, long_form=False):
    """
    Output some YAML
    """

    return yaml.dump(
        data,
        Dumper=get_dumper(clean_up, long_form),
        default_flow_style=False,
        allow_unicode=True
    )


def to_json(template, clean_up=False):
    """
    Assume the input is YAML and convert to JSON
    """

    data, _ = load(template)

    if clean_up:
        data = clean(data)

    return dump_json(data)


def to_yaml(template, clean_up=False, long_form=False, literal=True):
    """
    Assume the input is JSON and convert to YAML
    """

    data, _ = load(template)

    if clean_up:
        data = clean(data)

    if literal:
        data = cfn_literal_parser(data)

    return dump_yaml(data, clean_up, long_form)


def flip(template, in_format=None, out_format=None, clean_up=False, no_flip=False, long_form=False):
    """
    Figure out the input format and convert the data to the opposing output format
    """

    # Do we need to figure out the input format?
    if not in_format:
        # Load the template as JSON?
        if (out_format == "json" and no_flip) or (out_format == "yaml" and not no_flip):
            in_format = "json"
        elif (out_format == "yaml" and no_flip) or (out_format == "json" and not no_flip):
            in_format = "yaml"

    # Load the data
    if in_format == "json":
        data = load_json(template)
    elif in_format == "yaml":
        data = load_yaml(template)
    else:
        data, in_format = load(template)

    # Clean up?
    if clean_up:
        data = clean(data)

    # Figure out the output format
    if not out_format:
        if (in_format == "json" and no_flip) or (in_format == "yaml" and not no_flip):
            out_format = "json"
        else:
            out_format = "yaml"

    # Finished!
    if out_format == "json":
        if sys.version[0] == "3":
            return dump_json(data)
        else:
            return dump_json(data).encode('utf-8')

    return dump_yaml(data, clean_up, long_form)



"""
Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License").
You may not use this file except in compliance with the License. A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and limitations under the License.
"""

from cfn_tools.odict import ODict
from cfn_tools.literal import LiteralString

try:
    from collections.abc import KeysView
except ImportError:
    from collections import KeysView

import json
import six

UNCONVERTED_KEYS = [
    # Resource Type, String Attribute to keep Json
    ("AWS::StepFunctions::StateMachine", "DefinitionString")
]


def has_intrinsic_functions(parameter):
    intrinsic_functions = ["Fn::Sub", "!Sub", "!GetAtt"]
    result = False
    if isinstance(parameter, (list, tuple, dict, KeysView)):
        for item in parameter:
            if item in intrinsic_functions:
                result = True
                break
    return result


def convert_join(value):
    """
    Fix a Join ;)
    """

    if not isinstance(value, list) or len(value) != 2:
        # Cowardly refuse
        return value

    sep, parts = value[0], value[1]

    if isinstance(parts, six.string_types):
        return parts

    if not isinstance(parts, list):
        # This looks tricky, just return the join as it was
        return {
            "Fn::Join": value,
        }

    plain_string = True

    args = ODict()
    new_parts = []

    for part in parts:
        part = clean(part)

        if isinstance(part, dict):
            plain_string = False

            if "Ref" in part:
                new_parts.append("${{{}}}".format(part["Ref"]))
            elif "Fn::GetAtt" in part:
                params = part["Fn::GetAtt"]
                new_parts.append("${{{}}}".format(".".join(params)))
            else:
                for key, val in args.items():
                    # we want to bail if a conditional can evaluate to AWS::NoValue
                    if isinstance(val, dict):
                        if "Fn::If" in val and "AWS::NoValue" in str(val["Fn::If"]):
                            return {
                                "Fn::Join": value,
                            }

                    if val == part:
                        param_name = key
                        break
                else:
                    param_name = "Param{}".format(len(args) + 1)
                    args[param_name] = part

                new_parts.append("${{{}}}".format(param_name))

        elif isinstance(part, six.string_types):
            new_parts.append(part.replace("${", "${!"))

        else:
            # Doing something weird; refuse
            return {
                "Fn::Join": value
            }

    source = sep.join(new_parts)

    if plain_string:
        return source

    if args:
        return ODict((
            ("Fn::Sub", [source, args]),
        ))

    return ODict((
        ("Fn::Sub", source),
    ))


def clean(source):
    """
    Clean up the source:
    * Replace use of Fn::Join with Fn::Sub
    * Keep json body for specific resource properties
    """

    if isinstance(source, dict):
        for key, value in source.items():
            if key == "Fn::Join":
                return convert_join(value)

            else:
                source[key] = clean(value)

    elif isinstance(source, list):
        return [clean(item) for item in source]

    return source


def cfn_literal_parser(source):
    """
    Sanitize the source:
    * Keep json body for specific resource properties
    """

    if isinstance(source, dict):
        for key, value in source.items():
            if key == "Type":
                for item in UNCONVERTED_KEYS:
                    if value == item[0]:
                        # Checking if this resource has "Properties" and the property literal to maintain
                        # Better check than just try/except KeyError :-)
                        if source.get("Properties") and source.get("Properties", {}).get(item[1]):
                            if not has_intrinsic_functions(source["Properties"][item[1]].keys()):
                                source["Properties"][item[1]] = LiteralString(u"{}".format(json.dumps(
                                    source["Properties"][item[1]],
                                    indent=2,
                                    separators=(',', ': '))
                                ))

            else:
                source[key] = cfn_literal_parser(value)

    elif isinstance(source, list):
        return [cfn_literal_parser(item) for item in source]

    return source



"""
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License").
You may not use this file except in compliance with the License. A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and limitations under the License.
"""

from .main import main

main()



"""
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License").
You may not use this file except in compliance with the License.
A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and limitations under the License.
"""

import six

from cfn_clean.yaml_dumper import CleanCfnYamlDumper
from cfn_tools.literal import LiteralString
from cfn_tools.odict import ODict
from cfn_tools.yaml_dumper import CfnYamlDumper

TAG_STR = "tag:yaml.org,2002:str"
TAG_MAP = "tag:yaml.org,2002:map"
CONVERTED_SUFFIXES = ["Ref", "Condition"]

FN_PREFIX = "Fn::"

# Maximum length of a string before switching to angle-bracket-style representation
STR_MAX_LENGTH_QUOTED = 200
# Maximum number of newlines a string can have before switching to pipe-style representation
STR_MAX_LINES_QUOTED = 10


class Dumper(CfnYamlDumper):
    """
    The standard dumper
    """


class CleanDumper(CleanCfnYamlDumper):
    """
    Cleans up strings
    """


class LongDumper(CfnYamlDumper):
    """
    Preserves long-form function syntax
    """


class LongCleanDumper(CleanCfnYamlDumper):
    """
    Preserves long-form function syntax
    """


def literal_unicode_representer(dumper, value):
    return dumper.represent_scalar(TAG_STR, value, style='|')


def string_representer(dumper, value):
    if sum(1 for nl in value if nl in ('\n', '\r')) >= STR_MAX_LINES_QUOTED:
        return dumper.represent_scalar(TAG_STR, value, style="|")

    if len(value) >= STR_MAX_LENGTH_QUOTED and '\n' not in value:
        return dumper.represent_scalar(TAG_STR, value, style=">")

    if value.startswith("0"):
        return dumper.represent_scalar(TAG_STR, value, style="'")

    return dumper.represent_scalar(TAG_STR, value)


def fn_representer(dumper, fn_name, value):
    tag = "!{}".format(fn_name)

    if tag == "!GetAtt" and isinstance(value, list):
        value = ".".join(value)

    if isinstance(value, list):
        return dumper.represent_sequence(tag, value)

    if isinstance(value, dict):
        return dumper.represent_mapping(tag, value)

    return dumper.represent_scalar(tag, value)


def map_representer(dumper, value):
    """
    Deal with !Ref style function format and OrderedDict
    """

    value = ODict(value.items())

    if len(value.keys()) == 1:
        key = list(value.keys())[0]

        if key in CONVERTED_SUFFIXES:
            return fn_representer(dumper, key, value[key])

        if key.startswith(FN_PREFIX):
            return fn_representer(dumper, key[4:], value[key])

    return dumper.represent_mapping(TAG_MAP, value, flow_style=False)


# Customise our dumpers
Dumper.add_representer(ODict, map_representer)
Dumper.add_representer(six.text_type, string_representer)
Dumper.add_representer(LiteralString, literal_unicode_representer)
CleanDumper.add_representer(LiteralString, literal_unicode_representer)
CleanDumper.add_representer(ODict, map_representer)


def get_dumper(clean_up=False, long_form=False):
    if clean_up:
        if long_form:
            return LongCleanDumper

        return CleanDumper

    if long_form:
        return LongDumper

    return Dumper
